use std::{collections::HashSet, sync::Arc};

use btleplug::{
    api::{Central, CentralState, Manager as _, Peripheral, ScanFilter},
    platform::{Adapter, Manager, PeripheralId},
};
use futures::StreamExt;
use tauri::{
    async_runtime::{self, JoinHandle},
    AppHandle, Emitter,
};

use crate::{cmd::PeripheralItem, errors, state};

async fn notify_peripheral_update(app_handle: Arc<AppHandle>, peirpheral: &PeripheralId) {
    let adapter = state::get_central_adapter().await;
    if let Some(adapter) = adapter {
        let peripheral = adapter.peripheral(peirpheral).await;
        if let Ok(peripheral) = peripheral {
            let properties = peripheral.properties().await.ok().flatten();
            let services = state::get_found_setvices(peirpheral).await;
            let item =
                PeripheralItem::new_with_cache(&peripheral, properties.as_ref(), services).await;
            app_handle.emit("peripheral_updated", item).unwrap();
        }
    }
}

pub async fn handle_bluetooth_events(
    app_handle: Arc<AppHandle>,
) -> Result<JoinHandle<()>, errors::AppError> {
    let adapter = state::get_central_adapter().await;
    let app = Arc::clone(&app_handle);
    if let Some(adapter) = adapter {
        let adapter = adapter.clone();
        Ok(async_runtime::spawn(async move {
            let mut event_stream = adapter.events().await.unwrap();
            while let Some(event) = event_stream.next().await {
                match event {
                    btleplug::api::CentralEvent::DeviceDiscovered(id)
                    | btleplug::api::CentralEvent::DeviceUpdated(id) => {
                        notify_peripheral_update(Arc::clone(&app_handle), &id).await;
                        state::update_found_peripherals(id.clone()).await;
                    }
                    btleplug::api::CentralEvent::DeviceConnected(id) => {
                        state::set_connected_peripheral(id).await;
                        app.emit("peripheral_connected", ()).unwrap();
                    }
                    btleplug::api::CentralEvent::DeviceDisconnected(_id) => {
                        state::clear_connected_peripheral().await;
                        app.emit("peripheral_disconnected", ()).unwrap();
                    }
                    btleplug::api::CentralEvent::ServicesAdvertisement { id, services } => {
                        let services = HashSet::from_iter(services);
                        state::update_found_services(id.clone(), services).await;
                        notify_peripheral_update(Arc::clone(&app_handle), &id).await;
                    }
                    _ => {}
                }
            }
        }))
    } else {
        app.emit("central_state_updated", ()).unwrap();
        Err(errors::AppError::BluetoothNotReady)
    }
}

pub async fn get_central_adapter() -> Result<Adapter, errors::AppError> {
    let manager = Manager::new()
        .await
        .map_err(|_| errors::AppError::BluetoothNotReady)?;
    let adapters = manager
        .adapters()
        .await
        .map_err(|_| errors::AppError::BluetoothAdapterNotFound)?;

    let mut found_adapter = None;
    for adpater in adapters {
        let state = adpater.adapter_state().await;
        if let Ok(state) = state {
            if state == CentralState::PoweredOn {
                found_adapter = Some(adpater);
                break;
            }
        }
    }

    found_adapter.ok_or(errors::AppError::NoAvailableBluetoothAdapter)
}

pub async fn start_scan(app_handle: Arc<AppHandle>) -> Result<(), errors::AppError> {
    let adapter = state::get_central_adapter().await;
    if let Some(adapter) = adapter {
        adapter
            .start_scan(ScanFilter::default())
            .await
            .map_err(|_| errors::AppError::UnableToStartScan)?;
        state::set_scanning(true).await;
        state::clear_found_peripherals().await;
        app_handle.emit("scanning_started", ()).unwrap();
        Ok(())
    } else {
        Err(errors::AppError::NoAvailableBluetoothAdapter)
    }
}

pub async fn stop_scan(app_handle: Arc<AppHandle>) -> Result<(), errors::AppError> {
    let adapter = state::get_central_adapter().await;
    if let Some(adapter) = adapter {
        adapter
            .stop_scan()
            .await
            .map_err(|_| errors::AppError::UnableToStopScan)?;
        state::set_scanning(false).await;
        app_handle.emit("scanning_stopped", ()).unwrap();
        Ok(())
    } else {
        Err(errors::AppError::NoAvailableBluetoothAdapter)
    }
}
